import os
import os.path
os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"
import sys
import subprocess

sys.path.insert(0, os.getcwd())

import gradio as gr
from PIL import Image
import torch
import uuid
import shutil
import json
import yaml
from slugify import slugify
from transformers import AutoProcessor, AutoModelForCausalLM
from gradio_logsview import LogsView, LogsViewRunner
from huggingface_hub import hf_hub_download

trash_paths = ['tmp/gradio', 'datasets', 'pip']

for trash_path in trash_paths:
    if os.path.exists(trash_path):
        shutil.rmtree(trash_path)

MAX_IMAGES = 150
print(
    """
███████╗██╗░░░░░██╗░░░██╗██╗░░██╗░██████╗░██╗░░░██╗███╗░░░███╗░░░░░░
██╔════╝██║░░░░░██║░░░██║╚██╗██╔╝██╔════╝░╚██╗░██╔╝████╗░████║░░░░░░
█████╗░░██║░░░░░██║░░░██║░╚███╔╝░██║░░██╗░░╚████╔╝░██╔████╔██║█████╗
██╔══╝░░██║░░░░░██║░░░██║░██╔██╗░██║░░╚██╗░░╚██╔╝░░██║╚██╔╝██║╚════╝
██║░░░░░███████╗╚██████╔╝██╔╝╚██╗╚██████╔╝░░░██║░░░██║░╚═╝░██║░░░░░░
╚═╝░░░░░╚══════╝░╚═════╝░╚═╝░░╚═╝░╚═════╝░░░░╚═╝░░░╚═╝░░░░░╚═╝░░░░░░
░█████╗░██╗███╗░░██╗███████╗████████╗░██████╗██████╗░
██╔══██╗██║████╗░██║██╔════╝╚══██╔══╝██╔════╝██╔══██╗
███████║██║██╔██╗██║█████╗░░░░░██║░░░╚█████╗░██║░░██║
██╔══██║██║██║╚████║██╔══╝░░░░░██║░░░░╚═══██╗██║░░██║
██║░░██║██║██║░╚███║███████╗░░░██║░░░██████╔╝██████╔╝
╚═╝░░╚═╝╚═╝╚═╝░░╚══╝╚══════╝░░░╚═╝░░░╚═════╝░╚═════╝░
""")

def load_captioning(uploaded_files, concept_sentence):
    uploaded_images = [file for file in uploaded_files if not file.endswith('.txt')]
    txt_files = [file for file in uploaded_files if file.endswith('.txt')]
    txt_files_dict = {os.path.splitext(os.path.basename(txt_file))[0]: txt_file for txt_file in txt_files}
    updates = []
    if len(uploaded_images) <= 1:
        raise gr.Error(
            "Загрузите не менее 2 изображений для обучения модели (идеальное количество при настройках по умолчанию — от 4 до 30)"
        )
    elif len(uploaded_images) > MAX_IMAGES:
        raise gr.Error(f"На данный момент для обучения разрешено использовать не более {MAX_IMAGES} изображений.")

    updates.append(gr.update(visible=True))

    for i in range(1, MAX_IMAGES + 1):
        visible = i <= len(uploaded_images)
        updates.append(gr.update(visible=visible))
        image_value = uploaded_images[i - 1] if visible else None
        updates.append(gr.update(value=image_value, visible=visible))
        corresponding_caption = False
        if(image_value):
            base_name = os.path.splitext(os.path.basename(image_value))[0]
            if base_name in txt_files_dict:
                with open(txt_files_dict[base_name], 'r') as file:
                    corresponding_caption = file.read()

        text_value = corresponding_caption if visible and corresponding_caption else concept_sentence if visible and concept_sentence else None
        updates.append(gr.update(value=text_value, visible=visible))

    updates.append(gr.update(visible=True))
    updates.append(gr.update(visible=True))

    return updates

def hide_captioning():
    return gr.update(visible=False), gr.update(visible=False)

def resize_image(image_path, output_path, size):
    with Image.open(image_path) as img:
        width, height = img.size
        if width < height:
            new_width = size
            new_height = int((size/width) * height)
        else:
            new_height = size
            new_width = int((size/height) * width)
        print(f"Изменение размера {image_path} : {new_width}x{new_height}")
        img_resized = img.resize((new_width, new_height), Image.Resampling.LANCZOS)
        img_resized.save(output_path)

def create_dataset(destination_folder, size, *inputs):
    print("Создание датасета")
    images = inputs[0]
    if not os.path.exists(destination_folder):
        os.makedirs(destination_folder)

    for index, image in enumerate(images):
        new_image_path = shutil.copy(image, destination_folder)
        resize_image(new_image_path, new_image_path, size)
        original_caption = inputs[index + 1]

        image_file_name = os.path.basename(new_image_path)
        caption_file_name = os.path.splitext(image_file_name)[0] + ".txt"
        caption_path = resolve_path_without_quotes(os.path.join(destination_folder, caption_file_name))
        print(f"Путь к изображению={new_image_path}, путь к подписи = {caption_path}, оригинальная подпись={original_caption}")
        with open(caption_path, 'w') as file:
            file.write(original_caption)

    print(f"Папка назначения {destination_folder}")
    return destination_folder


def run_captioning(images, concept_sentence, *captions):
    print(f"Запуск_подписывания")
    print(f"Концептуальное предложение {concept_sentence}")
    print(f"Подписи {captions}")
    device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"Устройство={device}")
    torch_dtype = torch.float16
    model = AutoModelForCausalLM.from_pretrained(
        "multimodalart/Florence-2-large-no-flash-attn", torch_dtype=torch_dtype, trust_remote_code=True
    ).to(device)
    processor = AutoProcessor.from_pretrained("multimodalart/Florence-2-large-no-flash-attn", trust_remote_code=True)

    captions = list(captions)
    for i, image_path in enumerate(images):
        print(captions[i])
        if isinstance(image_path, str):
            image = Image.open(image_path).convert("RGB")

        prompt = "<DETAILED_CAPTION>"
        inputs = processor(text=prompt, images=image, return_tensors="pt").to(device, torch_dtype)
        print(f"Вводы {inputs}")

        generated_ids = model.generate(
            input_ids=inputs["input_ids"], pixel_values=inputs["pixel_values"], max_new_tokens=1024, num_beams=3
        )
        print(f"сгенерированы_ids {generated_ids}")

        generated_text = processor.batch_decode(generated_ids, skip_special_tokens=False)[0]
        print(f"сгенерирован_text: {generated_text}")
        parsed_answer = processor.post_process_generation(
            generated_text, task=prompt, image_size=(image.width, image.height)
        )
        print(f"разобранный_ответ = {parsed_answer}")
        caption_text = parsed_answer["<DETAILED_CAPTION>"].replace("The image shows ", "")
        print(f"текст подписи = {caption_text}, концептуальное предложение={concept_sentence}")
        if concept_sentence:
            caption_text = f"{concept_sentence} {caption_text}"
        captions[i] = caption_text

        yield captions
    model.to("cpu")
    del model
    del processor
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

def recursive_update(d, u):
    for k, v in u.items():
        if isinstance(v, dict) and v:
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def resolve_path(p):
    current_dir = os.path.dirname(os.path.abspath(__file__))
    norm_path = os.path.normpath(os.path.join(current_dir, p))
    return f"\"{norm_path}\""
def resolve_path_without_quotes(p):
    current_dir = os.path.dirname(os.path.abspath(__file__))
    norm_path = os.path.normpath(os.path.join(current_dir, p))
    return norm_path

def gen_sh(
    output_name,
    resolution,
    seed,
    workers,
    learning_rate,
    network_dim,
    max_train_epochs,
    save_every_n_epochs,
    timestep_sampling,
    guidance_scale,
    vram,
    optim,
    network_train_unet_only,
):

    print(f"Параметры:\nNetwork dim= {network_dim},\nMax train epochs= {max_train_epochs},\nSave every n epochs {save_every_n_epochs},\nTimestep sampling= {timestep_sampling},\nGuidance_scale= {guidance_scale},\nVram= {vram},\nOptimizer= {optim},\nNetwork train unet only= {network_train_unet_only}\n")

    line_break = "\\"
    file_type = "sh"
    if sys.platform == "win32":
        line_break = "^"
        file_type = "bat"

    pretrained_model_path = resolve_path("models/unet/flux1-dev.sft")
    clip_path = resolve_path("models/clip/clip_l.safetensors")
    t5_path = resolve_path("models/clip/t5xxl_fp16.safetensors")
    ae_path = resolve_path("models/vae/ae.sft")
    output_dir = resolve_path("outputs")

    if optim == "adafactor":
        optimset ="--optimizer_type adafactor"

    elif optim == "adamw8bit":
        optimset ="--optimizer_type adamw8bit"

    if network_train_unet_only == "вкл":
        unet_only ="--network_train_unet_only"

    elif network_train_unet_only == "выкл":
        unet_only =""

    if vram == "16ГБ":
        optimizer =f"""{optimset} {line_break}
  --optimizer_args "relative_step=False" "scale_parameter=False" "warmup_init=False" {line_break}
  --lr_scheduler constant_with_warmup {line_break}
  {unet_only} {line_break}
  --max_grad_norm 0.0 {line_break}"""
    elif vram == "12ГБ":
        optimizer =f"""{optimset} {line_break}
  --optimizer_args "relative_step=False" "scale_parameter=False" "warmup_init=False" {line_break}
  --split_mode {line_break}
  --network_args "train_blocks=single" {line_break}
  --lr_scheduler constant_with_warmup {line_break}
  {unet_only} {line_break}
  --max_grad_norm 0.0 {line_break}"""
    else:
        optimizer = f"{optimset} {line_break}"

    sh = f"""accelerate launch {line_break}
  --mixed_precision bf16 {line_break}
  --num_cpu_threads_per_process 1 {line_break}
  sd-scripts/flux_train_network.py {line_break}
  --pretrained_model_name_or_path {pretrained_model_path} {line_break}
  --clip_l {clip_path} {line_break}
  --t5xxl {t5_path} {line_break}
  --ae {ae_path} {line_break}
  --cache_latents_to_disk {line_break}
  --save_model_as safetensors {line_break}
  --sdpa --persistent_data_loader_workers {line_break}
  --max_data_loader_n_workers {workers} {line_break}
  --seed {seed} {line_break}
  --gradient_checkpointing {line_break}
  --mixed_precision bf16 {line_break}
  --save_precision bf16 {line_break}
  --network_module networks.lora_flux {line_break}
  --network_dim {network_dim} {line_break}
  {optimizer}
  --learning_rate {learning_rate} {line_break}
  --cache_text_encoder_outputs {line_break}
  --cache_text_encoder_outputs_to_disk {line_break}
  --fp8_base {line_break}
  --highvram {line_break}
  --max_train_epochs {max_train_epochs} {line_break}
  --save_every_n_epochs {save_every_n_epochs} {line_break}
  --dataset_config {resolve_path('dataset.toml')} {line_break}
  --output_dir {output_dir} {line_break}
  --output_name {output_name} {line_break}
  --timestep_sampling {timestep_sampling} {line_break}
  --discrete_flow_shift 3.1582 {line_break}
  --model_prediction_type raw {line_break}
  --guidance_scale {guidance_scale} {line_break}
  {unet_only} {line_break}
  --loss_type l2 {line_break}"""
    return sh

def gen_toml(
  dataset_folder,
  resolution,
  class_tokens,
  num_repeats
):
    toml = f"""[general]
shuffle_caption = false
caption_extension = '.txt'
keep_tokens = 1

[[datasets]]
resolution = {resolution}
batch_size = 1
keep_tokens = 1

  [[datasets.subsets]]
  image_dir = '{resolve_path_without_quotes(dataset_folder)}'
  class_tokens = '{class_tokens}'
  num_repeats = {num_repeats}"""
    return toml

def update_total_steps(max_train_epochs, num_repeats, images):
    try:
        num_images = len(images)
        total_steps = max_train_epochs * num_images * num_repeats
        print(f"Кол-во эпох= {max_train_epochs},\nКол-во изображений= {num_images},\nКол-во повторов= {num_repeats},\nВсего шагов={total_steps}\n")
        return gr.update(value = total_steps)
    except:
        print("")


def start_training(
    train_script,
    train_config,
):
    os.makedirs("models", exist_ok=True)
    os.makedirs("outputs", exist_ok=True)

    file_type = "sh"
    if sys.platform == "win32":
        file_type = "bat"

    sh_filename = f"train.{file_type}"
    with open(sh_filename, 'w', encoding="utf-8") as file:
        file.write(train_script)
    gr.Info(f"Сгенерирован скрипт обучения {sh_filename}")

    with open('dataset.toml', 'w', encoding="utf-8") as file:
        file.write(train_config)
    gr.Info(f"Сгенерирован dataset.toml")

    if sys.platform == "win32":
        command = resolve_path_without_quotes('train.bat')
    else:
        command = f"bash {resolve_path('train.sh')}"

    env = os.environ.copy()
    env['PYTHONIOENCODING'] = 'utf-8'
    runner = LogsViewRunner()
    cwd = os.path.dirname(os.path.abspath(__file__))
    gr.Info(f"Запуск обучения")
    yield from runner.run_command([command], cwd=cwd)
    yield runner.log(f"Исполнитель: {runner}")
    gr.Info(f"Обучение завершено. Проверьте папку outputs на наличие файлов LoRA.", duration=None)

def update(
    lora_name,
    resolution,
    seed,
    workers,
    class_tokens,
    learning_rate,
    network_dim,
    max_train_epochs,
    save_every_n_epochs,
    timestep_sampling,
    guidance_scale,
    vram,
    optim,
    num_repeats,
    network_train_unet_only,
):
    output_name = slugify(lora_name)
    dataset_folder = str(f"datasets/{output_name}")
    sh = gen_sh(
        output_name,
        resolution,
        seed,
        workers,
        learning_rate,
        network_dim,
        max_train_epochs,
        save_every_n_epochs,
        timestep_sampling,
        guidance_scale,
        vram,
        optim,
        network_train_unet_only,
    )
    toml = gen_toml(
        dataset_folder,
        resolution,
        class_tokens,
        num_repeats
    )
    return gr.update(value=sh), gr.update(value=toml), dataset_folder

def loaded():
    print("FluxGym-AINetSD Запущен\n")

theme = gr.themes.Monochrome(
    text_size=gr.themes.Size(lg="18px", md="15px", sm="13px", xl="22px", xs="12px", xxl="24px", xxs="9px"),
    font=[gr.themes.GoogleFont("Source Sans Pro"), "ui-sans-serif", "system-ui", "sans-serif"],
)
css = """
@keyframes rotate {
    0% {
        transform: rotate(0deg);
    }
    100% {
        transform: rotate(360deg);
    }
}
h1{font-family: georgia; font-style: italic; font-weight: bold; font-size: 30px; letter-spacing: -1px;}
h3{margin-top: 0}
.tabitem{border: 0px}
.group_padding{}
nav{position: fixed; top: 0; left: 0; right: 0; z-index: 1000; text-align: center; padding: 10px; box-sizing: border-box; display: flex; align-items: center; backdrop-filter: blur(10px); }
nav button { background: none; color: firebrick; font-weight: bold; border: 2px solid firebrick; padding: 5px 10px; border-radius: 5px; font-size: 14px; }
nav img { height: 40px; width: 40px; border-radius: 40px; }
nav img.rotate { animation: rotate 2s linear infinite; }
.flexible { flex-grow: 1; }
.tast-details { margin: 10px 0 !important; }
.toast-wrap { bottom: var(--size-4) !important; top: auto !important; border: none !important; backdrop-filter: blur(10px); }
.toast-title, .toast-text, .toast-icon, .toast-close { color: black !important; font-size: 14px; }
.toast-body { border: none !important; }
#terminal { box-shadow: none !important; margin-bottom: 25px; background: rgba(0,0,0,0.03); }
#terminal .generating { border: none !important; }
#terminal label { position: absolute !important; }
#container { margin-top: 50px; }
.hidden { display: none !important; }
.codemirror-wrapper .cm-line { font-size: 12px !important; }
"""

js = """
function() {
    let autoscroll = document.querySelector("#autoscroll")
    if (window.iidxx) {
        window.clearInterval(window.iidxx);
    }
    window.iidxx = window.setInterval(function() {
        let text=document.querySelector(".codemirror-wrapper .cm-line").innerText.trim()
        let img = document.querySelector("#logo")
        if (text.length > 0) {
            autoscroll.classList.remove("hidden")
            if (autoscroll.classList.contains("on")) {
                autoscroll.textContent = "Autoscroll ON"
                window.scrollTo(0, document.body.scrollHeight, { behavior: "smooth" });
                img.classList.add("rotate")
            } else {
                autoscroll.textContent = "Autoscroll OFF"
                img.classList.remove("rotate")
            }
        }
    }, 500);
    console.log("autoscroll", autoscroll)
    autoscroll.addEventListener("click", (e) => {
        autoscroll.classList.toggle("on")
    })
}
"""

with gr.Blocks(elem_id="app", theme=theme, css=css, fill_width=True) as demo:
    output_components = []
    with gr.Row():
        gr.HTML("""<nav>
    <img id='logo' src='/file=icon.png' width='80' height='80'>
    <div class='flexible'></div>
    <button id='autoscroll' class='on hidden'></button>
</nav>
""")
    with gr.Row(elem_id='container'):
        with gr.Column():
            gr.Markdown(
    """# FluxGym-<a href="https://t.me/AINetSD" target="_blank">AINetSD</a>
            <p style="margin-top:0">Шаг 1. Настройте параметры обучения Лоры.</p>
            """, elem_classes="group_padding")

            lora_name = gr.Textbox(
                label="Имя вашей Лоры",
                info="Оно должно быть уникальным",
                placeholder="Например: Persian Miniature Painting style, Cat Toy",
            )
            concept_sentence = gr.Textbox(
                label="Триггерное слово/предложение",
                info="Слово-триггер или предложение, которое будет использоваться",
                placeholder="необычное слово, например p3rs0n или trtcrd, или предложение типа 'in the style of CNSTLL'",
                interactive=True,
            )
            vram = gr.Radio(["20ГБ", "16ГБ", "12ГБ" ], value="20ГБ", label="VRAM", interactive=True)
            optim = gr.Radio(["adafactor", "adamw8bit"], value="adafactor", label="ОПТИМИЗАТОР", interactive=True)
            num_repeats = gr.Number(value=10, precision=0, label="Repeat trains per image", interactive=True)
            max_train_epochs = gr.Number(label="Max Train Epochs", value=16, interactive=True)
            total_steps = gr.Number(0, interactive=False, label="Expected training steps")
            with gr.Accordion("Дополнительные настройки", open=False):
                seed = gr.Number(label="Seed", value=42, interactive=True)
                workers = gr.Number(label="Workers", value=2, interactive=True)
                learning_rate = gr.Textbox(label="Learning Rate", value="8e-4", interactive=True)
                save_every_n_epochs = gr.Number(label="Save every N epochs", value=4, interactive=True)
                guidance_scale = gr.Number(label="Guidance Scale", value=1.0, interactive=True)
                timestep_sampling = gr.Textbox(label="Timestep Sampling", value="shift", interactive=True)
                network_dim = gr.Number(label="LoRA Rank", value=4, minimum=4, maximum=128, step=4, interactive=True)
                network_train_unet_only = gr.Radio(["вкл", "выкл"], value="выкл", label="Train U-Net Only", interactive=True)
                resolution = gr.Radio([512, 768, 1024], value=512, label="Resize dataset images")
        with gr.Column():
            gr.Markdown(
                """# Шаг 2. Датасет
<p style="margin-top:0">Убедитесь, что подписи содержат триггерное слово.</p>
""", elem_classes="group_padding")
            images = gr.File(
                file_types=["image", ".txt"],
                label="Загрузите ваши изображения",
                file_count="multiple",
                interactive=True,
                visible=True,
                scale=1,
            )
            with gr.Group(visible=False) as captioning_area:
                do_captioning = gr.Button("Добавить AI-подписи с помощью Florence-2")
                output_components.append(captioning_area)
                caption_list = []
                for i in range(1, MAX_IMAGES + 1):
                    locals()[f"captioning_row_{i}"] = gr.Row(visible=False)
                    with locals()[f"captioning_row_{i}"]:
                        locals()[f"image_{i}"] = gr.Image(
                            type="filepath",
                            width=111,
                            height=111,
                            min_width=111,
                            interactive=False,
                            scale=2,
                            show_label=False,
                            show_share_button=False,
                            show_download_button=False,
                        )
                        locals()[f"caption_{i}"] = gr.Textbox(
                            label=f"Caption {i}", scale=15, interactive=True
                        )

                    output_components.append(locals()[f"captioning_row_{i}"])
                    output_components.append(locals()[f"image_{i}"])
                    output_components.append(locals()[f"caption_{i}"])
                    caption_list.append(locals()[f"caption_{i}"])
        with gr.Column():
            gr.Markdown(
                """# Шаг 3. Обучение
<p style="margin-top:0">Нажмите старт для начала обучения.</p>
""", elem_classes="group_padding")
            start = gr.Button("Старт", visible=True)
            output_components.append(start)
            with gr.Accordion("Скрипт обучения", open=False):
                train_script = gr.Textbox(label="Скрипт:", lines=40, interactive=True)
            with gr.Accordion("Конфиг обучения", open=False):
                train_config = gr.Textbox(label="Конфиг:", lines=15, interactive=True)

    with gr.Row():
        terminal = LogsView(label="Логи обучения", elem_id="terminal")

    dataset_folder = gr.State()
    listeners = [
        lora_name,
        resolution,
        seed,
        workers,
        concept_sentence,
        learning_rate,
        network_dim,
        max_train_epochs,
        save_every_n_epochs,
        timestep_sampling,
        guidance_scale,
        vram,
        optim,
        num_repeats,
        network_train_unet_only,
    ]


    for listener in listeners:
        listener.change(update, inputs=listeners, outputs=[train_script, train_config, dataset_folder])

    images.upload(
        load_captioning,
        inputs=[images, concept_sentence],
        outputs=output_components
    )

    images.delete(
        load_captioning,
        inputs=[images, concept_sentence],
        outputs=output_components
    )

    images.clear(
        hide_captioning,
        outputs=[captioning_area, start]
    )

    max_train_epochs.change(
        fn=update_total_steps,
        inputs=[max_train_epochs, num_repeats, images],
        outputs=[total_steps]
    )
    num_repeats.change(
        fn=update_total_steps,
        inputs=[max_train_epochs, num_repeats, images],
        outputs=[total_steps]
    )

    images.upload(
        fn=update_total_steps,
        inputs=[max_train_epochs, num_repeats, images],
        outputs=[total_steps]
    )
    images.delete(
        fn=update_total_steps,
        inputs=[max_train_epochs, num_repeats, images],
        outputs=[total_steps]
    )
    images.clear(
        fn=update_total_steps,
        inputs=[max_train_epochs, num_repeats, images],
        outputs=[total_steps]
    )

    start.click(fn=create_dataset, inputs=[dataset_folder, resolution, images] + caption_list, outputs=dataset_folder).then(
        fn=start_training,
        inputs=[
            train_script,
            train_config
        ],
        outputs=terminal,
    )

    do_captioning.click(fn=run_captioning, inputs=[images, concept_sentence] + caption_list, outputs=caption_list)
    demo.load(fn=loaded, js=js)

if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    demo.launch(inbrowser=True,show_error=True, allowed_paths=[cwd])